package com.wdkl.ncs.app.powercontroltest;

import static androidx.constraintlayout.motion.utils.Oscillator.TAG;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.DialogInterface;
import android.os.Bundle;
import android.serialport.SerialPort;
import android.util.Log;


import com.wdkl.ncs.app.powercontroltest.common.Constants;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.security.InvalidParameterException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.CRC32;

public abstract class SerialPortActivity extends Activity {

    protected Application mApplication;
    protected SerialPort mSerialPort;
    protected OutputStream mOutputStream;
    private InputStream mInputStream;
    private ReadThread mReadThread;

    private Map<Long, List<byte[]>> msgCache = new HashMap<>();

    private class ReadThread extends Thread {
        ByteBuffer byteBuffer = ByteBuffer.allocate(512);

        @Override
        public void run() {
            super.run();
            while (!isInterrupted()) {
                try {
                    Thread.sleep(200);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
                int size;
                try {

                    byte[] buffer = new byte[128];
                    if (mInputStream == null) return;
                    if (mInputStream.available() > 0) {
                        size = mInputStream.read(buffer);
                        Log.e(TAG, "run: buffer length:" + byteBuffer.position());
                        byteBuffer.put(ArrayUtils.subarray(buffer, 0, size));
                        String msg= handlePackage(byteBuffer);
                        if(StringUtils.isNotEmpty(msg)){
                            byte[] bytes = msg.getBytes();
                            onDataReceived(bytes, bytes.length);
                        }
//                        String msg=getMsg(byteBuffer);
//                        while (StringUtils.isNotEmpty(msg)){
//                            byte[] bytes = msg.getBytes();
//                            onDataReceived(bytes, bytes.length);
//                            Log.i("收到消息", "run: "+msg);
//                            byteBuffer.compact();
//                            msg=getMsg(byteBuffer);
//                        }
//                        byteBuffer.compact();
//                        if(){
//                            byte[] bytes = msg.getBytes();
//                            onDataReceived(bytes, bytes.length);
//                            Log.i("收到消息", "run: "+msg);
//                        }
//                        if (size > 0) {
//                            onDataReceived(buffer, size);
//                        }
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                    return;
                }
            }
        }

        private String handlePackage(ByteBuffer buffer) {
            try {
                buffer.flip();
                Log.e(TAG, "buff信息:position="+buffer.position()+",limit="+buffer.limit() );
                if (buffer.position() == 0 && buffer.limit() == 0) { //buffer是空的无需处理
                    buffer.compact();
                    return null;
                }
                for (int i = 0; i < buffer.limit() - 1; i++) {
                    //监测到包头
                    if (byteArrToHexString(new byte[]{buffer.get(i), buffer.get(i + 1)}).equalsIgnoreCase("A3B1")) {
                        if (i > 0) {
                            byte[] incorrect = new byte[i + 1];
                            buffer.get(incorrect, 0, i + 1);
                        }
                        //读取120个字节
                        if (buffer.limit() - i >= 120) {
                            byte[] bytes = new byte[120];
                            buffer.get(bytes, 0, 120);
                            Log.i(TAG, "handlePackage:  收到：" + ArrayUtils.toString(bytes));
                            //校验数据包
                            CRC32 crc32 = new CRC32();
                            byte[] subarray1 = ArrayUtils.subarray(bytes, 0, bytes.length - 8);
                            Log.i(TAG, "handlePackage: " + ArrayUtils.toString(subarray1));
                            crc32.update(subarray1);
                            //计算出的crc校验
                            long caculateCheck = crc32.getValue();
                            byte[] subarray = ArrayUtils.subarray(bytes, bytes.length - 8, bytes.length);
                            //传输过来的校验
                            long transCheck = ByteBuffer.wrap(subarray).getLong();
                            if (caculateCheck == transCheck) {
                                long msgId = ByteBuffer.wrap(ArrayUtils.subarray(bytes, 5, 13)).getLong();
                                int packageTotal = bytes[4];
                                int packageIndex = bytes[3];
                                int availableLength = bytes[2];
                                byte[] availalbeData = ArrayUtils.subarray(bytes, 13, 13 + availableLength);
                                if (packageTotal == 1) { //单包消息
                                    String s = new String(availalbeData);
                                    Log.i(TAG, "收到完整消息: " + s);
                                    buffer.compact();
                                    return s;

                                } else {
                                    if (!msgCache.containsKey(msgId)) {
                                        List<byte[]> list = new ArrayList<>(packageTotal);
                                        for (int j = 0; j < packageTotal; j++) {
                                            list.add(null);
                                        }
                                        list.set(packageIndex - 1, availalbeData);
                                        msgCache.put(msgId, list);
                                    } else {
                                        List<byte[]> exist = msgCache.get(msgId);
                                        exist.set(packageIndex - 1, availalbeData);
                                        if (exist.stream().allMatch(p->p!=null)) { //消息已经接收完整
                                            int sum = exist.stream().mapToInt(p -> p.length).sum();
                                            ByteBuffer combine = ByteBuffer.allocate(sum);
                                            for (byte[] bytes1 : exist) {
                                                combine.put(bytes1);
                                            }
                                            String s = new String(combine.array());
                                            msgCache.remove(msgId);
                                            Log.i(TAG, "收到完整消息: " + s);
                                            buffer.compact();
                                            return s;
                                        }

                                    }
                                    Log.i(TAG, "handlePackage: 消息ID：" + msgId + "crc-c:" + caculateCheck + ",crc-t:" + transCheck);
                                }
                            }

                        }
                        break;
                    }
                }

                if(buffer.limit()>=360){ //如果在3个数据包的缓冲区长度内都没有发现一个包头，清空buffer,防止其他信号干扰
                    buffer.get(new byte[buffer.limit()],0,buffer.limit());
                }
                buffer.compact();

                return null;
            } catch (Exception e) {
                e.printStackTrace();
            }
            return null;
        }

        private String getMsg(ByteBuffer buffer) {
            try {
                buffer.flip();
                if (buffer.position() == 0 && buffer.limit() == 0) { //buffer是空的无需处理
                    return null;
                }

                StringBuilder sb = new StringBuilder();
                //截取一段 AA55打头 直到下一个AA55 之前的数据，校验是否为一个可用的消息。如果是，返回，不是则丢弃
                int start = 0, end = 0;
                for (int i = 0; i < buffer.limit() - 1; i++) {
                    //去除帧头之前无用数据
                    if (byteArrToHexString(new byte[]{buffer.get(i), buffer.get(i + 1)}).equalsIgnoreCase("AA55")) {

                        if (start == 0) {
                            start = i;
                        } else if (end == 0) {
                            end = i;
                        }
                    }
                }
                if ((end >= 0 && end - start > 6) || end == 0) { //end =0 消息单包找不到第二个包头。end >0 找到了第二个包头
                    if (start > 0) {
                        byte[] incorrect = new byte[start + 1];
                        buffer.get(incorrect, 0, start + 1);
                    }

                    //帧头
                    byte[] bytes = {buffer.get(start), buffer.get(start + 1)};
                    //数据长度
                    byte[] bytesLength = {buffer.get(start + 2), buffer.get(start + 3), buffer.get(start + 4), buffer.get(start + 5)};
                    int length = ByteBuffer.wrap(bytesLength).getInt();
                    //消息实际长度
                    int realLength = buffer.limit() - buffer.position();
                    if (end > 0) {
                        realLength = end - start;
                    }
                    if (length != realLength) { //消息长度指示与获取到的数据实际长度不符丢弃消息
                        if (end > 0) { //找到了第二个包头，说明当前消息不会再有后续了，没有找到第二个包头，说明接收可能还未完
                            byte[] msg = new byte[realLength];
                            buffer.get(msg, 0, realLength);
                        }
                        return null;
                    }
                    byte[] msg = new byte[length];
                    buffer.get(msg, 0, length);
                    //校验数据包
                    CRC32 crc32 = new CRC32();
                    crc32.update(ArrayUtils.subarray(msg, 0, msg.length - 8));
                    //crc校验
                    long value = crc32.getValue();
                    Log.i(TAG, "收到消息计算出的crc" + value);
                    byte[] subarray = ArrayUtils.subarray(msg, msg.length - 8, msg.length);
                    long aLong = ByteBuffer.wrap(subarray).getLong();
                    Log.i(TAG, "消息体内CRC " + aLong);
                    if (value == aLong) { //校验成功
                        sb.append(new String(ArrayUtils.subarray(msg, 6, msg.length - 8)));
                    }


//                            if (byteArrToHexString(bytes).equalsIgnoreCase("AA55") && (buffer.limit()-buffer.position()) >= length) { //判断到帧头
//                                Log.e(TAG, "buffer position: "+buffer.position()+",buffer limit"+buffer.limit() );
//                                byte[] msg = new byte[length];
//                                buffer.get(msg, buffer.position(), length);
//
//                            }


//                        if (i > 0){
//                            byte[] incorrect = new byte[i];
//                        buffer.get(incorrect, buffer.position(), i);

                }

//                if (buffer.position() < buffer.limit() && buffer.limit() - buffer.position() >= 14) { //帧头+数据长度为6个字节
//                    //帧头
//                    byte[] bytes = {buffer.get(buffer.position()), buffer.get(buffer.position() + 1)};
//                    //数据长度
//                    byte[] bytesLength = {buffer.get(buffer.position() + 2), buffer.get(buffer.position() + 3), buffer.get(buffer.position() + 4), buffer.get(buffer.position() + 5)};
//                    int length = ByteBuffer.wrap(bytesLength).getInt();
//                    if (byteArrToHexString(bytes).equalsIgnoreCase("AA55") && (buffer.limit()-buffer.position()) >= length) { //判断到帧头
//                        Log.e(TAG, "buffer position: "+buffer.position()+",buffer limit"+buffer.limit() );
//                        byte[] msg = new byte[length];
//                        buffer.get(msg, buffer.position(), length);
//                        //校验数据包
//                        CRC32 crc32 = new CRC32();
//                        crc32.update(ArrayUtils.subarray(msg, 0, msg.length-8));
//                        //crc校验
//                        long value = crc32.getValue();
//                        Log.i(TAG, "收到消息计算出的crc"+value);
//                        byte[] subarray = ArrayUtils.subarray(msg, msg.length - 8, msg.length);
//                        long aLong = ByteBuffer.wrap(subarray).getLong();
//                        Log.i(TAG, "消息体内CRC "+aLong);
//                        if(value==aLong) { //校验成功
//                            sb.append(new String(ArrayUtils.subarray(msg, 6, msg.length-8)));
//                        }
//                    }
//                }

                return sb.toString();
            } catch (Exception e) {
                e.printStackTrace();
            }
            return null;
        }

        /**
         * 字节数组转16进制字符串
         *
         * @param b
         * @return
         */
        private String byteArrToHexString(byte[] b) {
            String result = "";
            for (int i = 0; i < b.length; i++) {
                result += Integer.toString((b[i] & 0xff) + 0x100, 16).substring(1);
            }
            return result;
        }
    }

    private void DisplayError(int resourceId) {
        AlertDialog.Builder b = new AlertDialog.Builder(this);
        b.setTitle("Error");
        b.setMessage(resourceId);
        b.setPositiveButton("OK", new DialogInterface.OnClickListener() {
            public void onClick(DialogInterface dialog, int which) {
//                startActivity(new Intent(SerialPortActivity.this, SettingsActivity.class));
//                SerialPortActivity.this.finish();
            }
        });
        b.show();
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        mApplication = (Application) getApplication();

    }

    protected void openSerial() {
        try {
            if (mReadThread != null) mReadThread.interrupt();
            mApplication.closeSerialPort();
            mSerialPort = null;
            mSerialPort = mApplication.getSerialPort();
            mOutputStream = mSerialPort.getOutputStream();
            mInputStream = mSerialPort.getInputStream();

            /* Create a receiving thread */
            mReadThread = new ReadThread();
            mReadThread.start();
            Constants.SERIAL_PORT_STATUS = true;


        } catch (SecurityException e) {
            Constants.SERIAL_PORT_STATUS = false;
            Constants.SERIAL_PORT_FAIL_RESON = getResources().getString(R.string.error_security);
            DisplayError(R.string.error_security);
        } catch (IOException e) {
            Constants.SERIAL_PORT_STATUS = false;
            Constants.SERIAL_PORT_FAIL_RESON = getResources().getString(R.string.error_unknown);
            DisplayError(R.string.error_unknown);
        } catch (InvalidParameterException e) {
            Constants.SERIAL_PORT_STATUS = false;
            Constants.SERIAL_PORT_FAIL_RESON = getResources().getString(R.string.error_configuration);
            DisplayError(R.string.error_configuration);
        }
    }


    protected abstract void onDataReceived(final byte[] buffer, final int size);

    @Override
    protected void onDestroy() {
        if (mReadThread != null) mReadThread.interrupt();
        mApplication.closeSerialPort();
        mSerialPort = null;
        super.onDestroy();
    }
}